
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include <ctype.h>
#include "xstd/h/strstream.h"
#include "xstd/h/iomanip.h"

#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "base/polyLogCats.h"
#include "runtime/Connection.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/PubWorld.h"
#include "runtime/HttpDate.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/StatPhaseSync.h"
#include "runtime/SharedOpts.h"
#include "runtime/ConnMgr.h"
#include "runtime/ErrorMgr.h"
#include "runtime/LogComment.h"
#include "runtime/httpText.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentCfg.h"
#include "csm/ContentMgr.h"
#include "csm/oid2Url.h"
#include "client/CltOpts.h"
#include "client/Client.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/CltDataFilterRegistry.h"
#include "client/ChunkedCodingParser.h"
#include "client/AnyBodyParser.h"
#include "client/MarkupBodyParser.h"
#include "client/UriScriptBodyParser.h"
#include "client/ParseBuffer.h"
#include "client/RegExGroups.h"
#include "client/MembershipMap.h"
#include "client/AclGroup.h"
#include "client/CltCfg.h"
#include "client/CltXact.h"


CltXact::CltXact() {
	CltXact::reset();
}

void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theSrvRep = 0;
	theContentCfg = 0;
	theBodyParser = 0;
	theNextHop = NetAddr();
	theRepHdr.reset();
	theOlcTimes.reset();
	theReqBodySize = -1;
	thePeerState = peerUnknown;
	theCred.reset();
	theSavedRepHeader.reset();
	theCause = 0;
	theChildCount = 0;
	doRetry = false;

	theLogCat = lgcCltSide;
}

Agent *CltXact::owner() {
	return theOwner;
}

const UniqId &CltXact::reqId() const {
	return theId;
}

void CltXact::enqueue() {
	theEnqueTime = TheClock;
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(theOid);
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);
	if (!theOid.foreignUrl()) {
		theSrvRep = TheHostMap->serverRepAt(theOid.viserv());
		theContentCfg = TheContentMgr.get(theOid.type());
	}
	theOwner->cfg()->selectAbortCoord(theAbortCoord);

	theConn->theWr.start(this);

	Should (!theConn->tunneling() || theConn->sslConfigured());
	newState(theConn->tunneling() && !theConn->sslActive() ? 
		stTunnelConnWaiting : stConnWaiting);
}


void CltXact::finish(Error err) {
	if (!err && theRepHdr.redirect())
		redirect();

	if (!theOid.head()) {
		if (!err && !theOid.aborted() && theRepHdr.theChecksum.set()) {
			theCheckAlg.final();
			if (!theCheckAlg.sum().equal(theRepHdr.theChecksum)) {
				if (ReportError(errChecksum)) {
					theRepHdr.theChecksum.print(Comment << "MD5 expected: ") << endc;
					theCheckAlg.sum().print(Comment << "MD5 received: ") << endc;
					if (TheOpts.theDumpFlags(dumpErr, dumpAny))
						printMsg(theConn->theRdBuf);
				}
				err = errOther;
			}
		}		

		// cache objects with bodies
		if (!err && theOid.cachable() && theRepHdr.theContSize >= 0) {
			if (PrivCache *cache = theOwner->privCache())
				cache->storeOid(theOid);
		}
	}

	if (err && theOwner->privCache())
		theOwner->privCache()->purgeOid(theOid);

	Xaction::finish(err);
	if (theBodyParser) {
		theBodyParser->farm().put(theBodyParser);
		theBodyParser = 0;
	}

	theOwner->noteXactDone(this);
}

void CltXact::noteAbort() {
	newState(stDone);
}

void CltXact::noteReadReady(int) {
	if (!Should(theConn))
		return;

	if (abortIo(&Connection::read))
		return;

	if (!Should(theConn))
		return;

	const bool hdrWaiting = 
		theState == stHdrWaiting || theState == stTunnelRespWaiting;

	if (theConn->bad()) {
		if (!doRetry) {
			doRetry = hdrWaiting &&
				!theConn->theRdBuf.contSize() && theConn->useCnt() > 1;
		}
		finish(errOther);
		return;
	}

	if (hdrWaiting) {
		if (const Error err = getHeader()) {
			finish(err);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		getBody();
		return;
	}
}

Error CltXact::getHeader() {
	Assert(theState == stHdrWaiting || theState == stTunnelRespWaiting);

	if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		// change state immediately if we just established a tunnel
		if (theState == stTunnelRespWaiting) {
			if (theRepHdr.theStatus == RepHdr::sc200_OK) {

				static int connRespCount = 0;
				if (!connRespCount++ || TheOpts.theDumpFlags(dumpRep, dumpHdr))
					printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);

				theConn->theRdBuf.consumed(theRepHdr.theHdrSize);
				theRepHdr.reset();
				theConn->theRd.stop(this);
				theConn->theWr.start(this);
				newState(stConnWaiting);
				return 0;
			} else {
				// assume the normal code will handle the error
				newState(stHdrWaiting); // XXX: not last call in this function
				// continue
			}
		}

		const Error err = interpretHeader();
		// dump reply header after interpretHeader() to dump more oid flags
		static int respCount = 0;
		if (!respCount++ || TheOpts.theDumpFlags(dumpRep, dumpHdr))
			printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);

		if (err)
			return err;

		if (theHttpStatus == RepHdr::sc407_ProxyAuthRequired)
			saveRepHeader();

		consume(theRepHdr.theHdrSize);
		Assert(theState == stBodyWaiting);
		return err;
	} else
	if (expectMore()) {
		if (theConn->theRdBuf.full())
			return errHugeHdr;
		else
			return 0;
	} else {
		const bool readNothing = !theConn->theRdBuf.contSize();
		// HTTP pconn race condition?
		const bool race = readNothing && theConn->useCnt() > 1;
		doRetry = doRetry || race;
		if (race)
			return errOther;
		else
			return readNothing ? errNoHdrClose : errPrematureEoh;
	}
	Should(false); // not reached
	return errOther;
}

void CltXact::getBody() {
	Assert(theState == stBodyWaiting);
	Assert(theBodyParser);

	parse();

	if (expectMore()) {
		checkOverflow();
		return;
	}

	const Size leftoverSize = unconsumed();

	// no more data, set expected size if it was unknown
	if (!theRepSize.expected().known())
		theRepSize.expect(theRepSize.actual() + leftoverSize);

	const bool pgAborted = cfgAbortedReply();

	// make sure we do not leave anything behind and complain if needed
	if (leftoverSize > 0) {
		if (pgAborted) {
			// no need to complain, a Polygraph-initiated abort
		} else
		if (theRepSize.actual() + leftoverSize < theRepSize.expected()) {
			// no need for parser to complain since it is not a content error
		} else {
			const ParseBuffer leftovers(theConn->theRdBuf.content(),
				leftoverSize);
			theBodyParser->noteLeftovers(leftovers);
		}
		consume(leftoverSize);
	}

	authWithProxy(theHttpStatus == RepHdr::sc407_ProxyAuthRequired);

	if (theRepSize.expected() == theRepSize.actual()) {
		finish(0);
		return;
	}

	if (theRepSize.expected() < theRepSize.actual()) {
		finish(errExtraRepData);
		return;
	}

	Assert(theRepSize.expected() > theRepSize.actual());

	if (pgAborted) {
		theOid.aborted(true);
		finish(0); // not an error, configuration told server to abort
		return;
	}

	finish(errPrematureEof);
}

void CltXact::parse() {
	IOBuf &buf = theConn->theRdBuf;
	const Size availSize = unconsumed();
	const ParseBuffer data = availSize > 0 ?
		ParseBuffer(buf.content(), availSize) : ParseBuffer();

	if (const Size parsedSz = theBodyParser->parse(data)) {
		consume(parsedSz);
		theConn->theRdBuf.pack();
	}

	if (theRepSize.expectingWhatParsed()) {
		Should(!theRepSize.expected().known());
		theRepSize.expect(theRepSize.actual());
	}
}

void CltXact::checkOverflow() {
	const IOBuf &buf = theConn->theRdBuf;
	if (buf.contSize() >= buf.capacity()) { // full() requires pack()!
		const Size leftoverSize = unconsumed();
		const ParseBuffer leftovers(buf.content(), leftoverSize);
		theBodyParser->noteOverflow(leftovers);
		// if needed, we can let the parser decide how much to chop off
		consume(leftoverSize);
	}
}

	
void CltXact::noteWriteReady(int) {
	WrBuf &buf = theConn->theWrBuf;

	// perform initializations after connection is established
	if (theState == stTunnelConnWaiting) {
		makeConnectReq(theConn->theWrBuf);
		newState(stTunnelSpaceWaiting);
	} else
	if (theState == stConnWaiting) {
		if (theConn->sslConfigured() && !theConn->sslActive())
			theConn->sslActivate();
		makeExplicitReq(theConn->theWrBuf);
		newState(stSpaceWaiting);
	}

	Assert(buf.contSize() > 0);

	if (abortIo(&Connection::write))
		return;

	if (theConn->bad()) {
		Assert(theReqSize >= 0);
		// HTTP pconn race condition?
		const bool race = buf.contSize() == theReqSize && theConn->useCnt() > 1;
		doRetry = doRetry || race;
		finish(errOther);
		return;
	}

	if (buf.empty()) {
		theConn->theWr.stop(this);
		theConn->theRd.start(this);
		TheFileScanner->setPriority(theConn->sock().fd(), fsupBestEffort);
		if (theState == stSpaceWaiting && theOid.foreignUrl())
			TheEmbedStats.foreignUrlRequested++;
		newState(theState == stTunnelSpaceWaiting ?
			stTunnelRespWaiting : stHdrWaiting);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

// make a CONNECT request
void CltXact::makeConnectReq(WrBuf &buf) {
	static int reqCount = 0;
	ostrstream os(buf.space(), buf.spaceSize());

	os << rlpConnect;
	makeReqHost(os);
	makeReqVersion(os);
	os << crlf;

	finishReqHdrs(buf, os.pcount(), !reqCount++);

	// no body for CONNECT requests
}

// make a non-CONNECT request
void CltXact::makeExplicitReq(WrBuf &buf) {
	// decide whether the request should have a body
	if (theOid.post() || theOid.put()) {
		static RndGen rng;
		// try to make sure we will have space after the headers
		// may not work for small buffers (default 16KB is OK)
		theReqBodySize = rng(0, buf.spaceSize() / 2);
	}

	// make headers
	{
		static int reqCount = 0;
		ostrstream os(buf.space(), buf.spaceSize());
		makeReqMethod(os);
		makeEndToEndHdrs(os);
		makeHopByHopHdrs(os);
		finishReqHdrs(buf, os.pcount(), !reqCount++);
	}

	// where we should abort
	theAbortSize = theAbortCoord.pos(theReqSize, theReqBodySize);

	// append request body if needed
	if (theReqBodySize >= 0) {
		const char *bodyStart = buf.space();
		theReqSize += theReqBodySize;
		buf.appended(theReqBodySize); // garbage, for now

		if (TheOpts.theDumpFlags(dumpReq, dumpBody))
			printMsg(bodyStart, theReqBodySize);
	}
}

void CltXact::finishReqHdrs(WrBuf &buf, Size pfxSize, bool forceDump) {
	const char *reqStart = buf.space();
	buf.appended(pfxSize);

	// give filters a chance, they may add their own headers
	TheCltDataFilterRegistry().apply(this, buf);

	buf.append("\r\n", 2); // end-of-header

	// may be adjusted in makeExplicitReq() if there is a body
	theReqSize = buf.space() - reqStart;

	// dump request header
	if (forceDump || TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(reqStart, theReqSize);
}

void CltXact::makeReqMethod(ostream &os) {
	if (theOid.get())
		os << rlpGet;
	else
	if (theOid.post())
		os << rlpPost;
	else
	if (theOid.head())
		os << rlpHead;
	else
	if (theOid.put())
		os << rlpPut;
	else
		Assert(false);
}

void CltXact::makeReqVersion(ostream &os) {
	if (theOwner->httpVersion() <= HttpVersion(1,0))
		os << rlsHttp1p0;
	else
		os << rlsHttp1p1;
}

void CltXact::makeEndToEndHdrs(ostream &os) {

	if (theOwner->proxy()) { 
		// proxied connection, use full URL
		Oid2Url(theOid, os);
	} else {
		// transparent folks get an advantage of shorter request-lines
		Oid2UrlPath(theOid, os);
	}

	makeReqVersion(os);
	
	/* request-header fields */

	os << hfAccept;

	os << hfpHost;
	const HostCfg *host = makeReqHost(os);
	os << crlf;

	if (theOid.imsAny() && olcTimes().lmt() >= 0) {
		const Time t = theOid.ims200() ?
			(olcTimes().lmt() - Time::Sec(1)) : 
			(olcTimes().lmt() + Time::Sec(1));
		HttpDatePrint(os << hfpIMS, t) << crlf;
	}

	if (theOid.reload())
		os << hfReload;

	/* entity-header fields */

	os << hfpXXact << TheGroupId << ' ' << theId << crlf;

	// send public world info
	if (!theOid.foreignUrl()) {
		const PubWorld &pubWorld = *host->thePubWorld;
		os << hfpXLocWorld << pubWorld.localSlice() << crlf;

		if (const PubWorldSlice *slice = pubWorld.sliceToSync())
			os << hfpXRemWorld << *slice << crlf;

		if (theSrvRep)
			os << hfpXTarget << theSrvRep->addr() << crlf;
	}

	os << hfpXAbort << theAbortCoord.whether() 
		<< ' ' << theAbortCoord.where() << crlf;

	// report our readiness to change phase
	os << hfpXPhaseSyncPos << TheStatPhaseMgr.phaseSyncPos() << crlf;

	if (theReqBodySize >= 0)
		os << "Content-length: " << (int)theReqBodySize << crlf;
}

void CltXact::makeHopByHopHdrs(ostream &os) {
	/* general-header fields */

	// persistency indication depends on HTTP version
	if (theOwner->httpVersion() <= HttpVersion(1,0)) {
		if (theConn->reusable())
			os << (theOwner->proxy() ? hfConnAlivePxy : hfConnAliveOrg);
	} else {
		if (!theConn->reusable())
			os << (theOwner->proxy() ? hfConnClosePxy : hfConnCloseOrg);
	}

	if (theOwner->credsForProxy(theOid, theCred))
		PrintBase64(os << hfpProxyAuthorization, theCred.image().data(), theCred.image().len()) << crlf;
}

const HostCfg *CltXact::makeReqHost(ostream &os) {
	if (theOid.foreignUrl()) {
		const char *furi = theOid.foreignUrl().cstr();
		NetAddr addr;
		if (MsgHdr::ParseHostInUri(furi, furi+theOid.foreignUrl().len(), addr))
			os << addr.addrA() << ':' << addr.port();
		return 0;
	} else 
	if (const HostCfg *host = TheHostMap->at(theOid.viserv())) {
		const NetAddr &visName = host->theAddr;
		os << visName.addrA() << ':' << visName.port();
		return host;
	}
	Should(false);
	return 0;
}

void CltXact::consume(Size size) {
	theConn->theRdBuf.consumed(size);
	theRepSize.got(size);
}

Error CltXact::interpretHeader() {
	// does reply line make sense?
	if (!theRepHdr.theHttpVersion.known() || theRepHdr.theStatus < 0)
		return errHttpRLine;

	if (const Error err = setStatusCode(theRepHdr.theStatus))
		return err;

	// check content-length, set RepSize if possible
	if (theRepHdr.expectBody()) {
		if (theOid.head())
			theRepSize.expect(theRepHdr.theHdrSize);
		else
		if (theRepHdr.theTransferEncoding == MsgHdr::tcChunked) {
			if (theRepHdr.theContSize >= 0) // MUST ignore
				ReportError(errChunkedButCLen);
		} else
		if (theRepHdr.theContSize >= 0)
			theRepSize.expect(theRepHdr.theHdrSize + theRepHdr.theContSize);
		else
		if (theRepHdr.persistentConnection())
			return errPersistButNoCLen;
	} else {
		if (theRepHdr.theContSize >= 0)
			return errUnexpectedCLen;
		theRepSize.expect(theRepHdr.theHdrSize);
	}

	// check that we can handle transfer encoding
	if (theRepHdr.theTransferEncoding == MsgHdr::tcOther)
		return errUnknownTransferEncoding;

	theOid.repToRedir(theRepHdr.redirect());

	if (!theRepHdr.polyHeaders() && theRepHdr.expectPolyHeaders()) {
		theOid.foreignSrc(true);
		noteError(errForeignSrc);
	}

	if (theOid.foreignUrl())
		TheEmbedStats.foreignUrlReceived++;

	checkAcl();

	// firstHand here means our request reached the server and
	// server's reply reached us
	const bool firstHand = reqId().myMutant(theRepHdr.theXactId);

	theOid.cachable(theRepHdr.isCachable);
	theOid.hit(!firstHand && theRepHdr.theXactId && theOid.basic());

	if (theOid.hit()) {
		if (!theOid.offeredHit() && !TheCltOpts.ignoreFalseHits)
			noteError(errFalseHit);

		if (!theOid.cachable())
			noteError(errUnchbHit);
	}

	if (firstHand) { // native miss
		checkDateSync();
		firstHandSync();
	} else
	if (theRepHdr.theXactId) { // native second-hand response
		if (theOid.reload())
			noteError(errReloadHit);
	}

	if (theOid.offeredHit() && !theOid.hit())
		noteError(errFalseMiss);

	checkFreshness();

	// note if the server will close the connection
	theConn->lastUse(!theRepHdr.persistentConnection());

	theBodyParser = selectBodyParser();

	newState(stBodyWaiting);
	return 0;
}

void CltXact::noteError(const Error &err) {
	if (err == errFalseMiss) {
		if (TheCltOpts.printFalseMisses) {
			Oid2Url(theOid, cout << "False-Miss: ");
			cout << endl;
			printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);
		}
		return; // do not report false misses; they may not be errors
	}

	if (err == errForeignSrc) {
		if (theOid.foreignUrl())
			return; // response to a foreign URL may be foreign

		static bool informed = false;
		if (!informed) {
			Comment(5) << "fyi: received first foreign response to " <<
				"Polygraph-specific URL:" << endc;
			printMsg(theConn->theRdBuf);
			informed = true;
		}

		if (theHttpStatus == RepHdr::sc200_OK && theOwner->cfg()->acl())
			return; // let checkAcl() handle this case

		if (TheOpts.acceptForeignMsgs)
			return; // the user told us to accept all foreign responses
	}

	if (!ReportError(err))
		return;

	/* supply additional information */

	if (err == errStaleHit) {
		HttpDatePrint(cout << "\trep: ", theRepHdr.theDate) << endl;
		HttpDatePrint(cout << "\tlmt: ", olcTimes().lmt()) << endl;
		HttpDatePrint(cout << "\treq: ", theStartTime) << endl;
		HttpDatePrint(cout << "\tnow: " ) << endl;
		HttpDatePrint(cout << "\texp: ", olcTimes().exp()) << endl;
	}

	if (TheOpts.theDumpFlags(dumpErr, dumpAny))
		printMsg(theConn->theRdBuf);
}

void CltXact::noteCredentialsGone() {
	theCred.reset();
}

void CltXact::saveRepHeader() {
	theConn->theRdBuf.copyContent(theSavedRepHeader, theRepHdr.theHdrSize);
}

void CltXact::firstHandSync() {
	// update public world info
	if (theRepHdr.theRemWorld)
		updatePubWorld(theRepHdr.theRemWorld);

	// sync oid accounting on first-hand replies
	if (theSrvRep)
		theSrvRep->syncSrvInfo(theRepHdr);

	// sync phases
	if (theRepHdr.thePhaseSyncPos >= 0 && theRepHdr.theGroupId)
		TheStatPhaseSync.notePhaseSync(theRepHdr.theGroupId, theRepHdr.thePhaseSyncPos);
}

void CltXact::authWithProxy(bool needed) {
	bool reportErrDetails = false;
	if (needed) { // denied
		if (!theCred.image()) { // sent nothing
			doRetry = true;
			theOwner->noteProxyAuthReq(this);
		} else
		if (theCred.valid()) { // sent valid
			reportErrDetails = ReportError(errProxyAuthDenied);
		}
	} else {      // allowed
		if (theCred.image() && !theCred.valid()) // sent invalid
			reportErrDetails = ReportError(errProxyAuthAllowed);
	}

	if (reportErrDetails) {
		Comment << "robot: " << theOwner->host()
			<< " credentials: " << theCred.image()
			<< " proxy: " << theOwner->proxy() << endc;
	}
}

void CltXact::redirect() {
	ObjId dest = theRepHdr.theLocn.oid;
	dest.rediredReq(true);
	dest.get(true);

	if (!dest.foreignUrl()) {
		if (!validRelOid(dest)) {
			ReportError(errRedirLocation);
			return;
		}

		if (theRepHdr.theLocn.host) {
			if (const Error err = setViserv(theRepHdr.theLocn.host, dest)) {
				noteError(err);
				return;
			}
		} else {
			dest.viserv(theOid.viserv());
		}
	
		theOwner->selectTarget(dest);
		dest.repeat(true); // polysrv only redirects to seen URLs
	} else {
		dest.repeat(false); // XXX: may be a repeat, we do not know
	}
	// XXX: hot() is unset

	theOwner->noteRedirect(this, dest);
}

BodyParser *CltXact::selectBodyParser() {
	BodyParser *cparser = selectContentParser();

	if (theRepHdr.chunkedEncoding())
		return ChunkedCodingParser::GetOne(this, cparser);

	return cparser;
}

BodyParser *CltXact::selectContentParser() {
	// parse if content is markup
	if (theRepHdr.markupContent())
		return selectMarkupBodyParser();
	
	// if content is unknown, use URL extension (if any) to guess type
	if (!theRepHdr.knownContentType() && OidImpliesMarkup(theOid, theContentCfg))
		return selectMarkupBodyParser();

	// parse if domestic content may have embedded tags
	if (!theOid.foreignUrl() && theContentCfg->hasEmbedCont())
		return selectMarkupBodyParser();

	// do not parse otherwise
	return AnyBodyParser::GetOne(this);
}

BodyParser *CltXact::selectMarkupBodyParser() {
	if (theOwner->cfg()->followAllUris(theRepHdr))
			return UriScriptBodyParser::GetOne(this, theOwner->cfg());
	return MarkupBodyParser::GetOne(this, theOwner->cfg());
}

bool CltXact::expectMore() const {
	if (theConn->atEof())
		return false;

	if (!theRepSize.expected().known())
		return true;

	return theRepSize.expected() > theRepSize.actual() + unconsumed();
}

Size CltXact::unconsumed() const {
	return theRepSize.expectToGet(theConn->theRdBuf.contSize());
}

const ObjTimes &CltXact::olcTimes() const {
	// not calculated or not configured, but can be
	if (theOlcTimes.lmt() < 0 && theContentCfg) 
		theContentCfg->calcTimes(theOid, theOlcTimes);
	return theOlcTimes;
}

bool CltXact::validRelOid(const ObjId &oid) const {
	return oid.world() && oid.name() > 0 && 
		TheContentMgr.validId(oid.type());
}

void CltXact::checkAcl() {
	const AclGroup &acl = theOwner->cfg()->acl();
	if (!acl) // no access controls configured
		return;

	if (!acl.needsCheck(theOid.foreignUrl()))
		return;

	if (RepHdr::PositiveStatusCode(theHttpStatus)) {
		if (!theOid.foreignUrl() && theOid.foreignSrc())
			checkAclMatch(acl.rewrite(), "rewritten");
		else
			checkAclMatch(acl.allow(), "allowed");
		return;
	} 

	if (theHttpStatus == RepHdr::sc407_ProxyAuthRequired) {
		// do not check acls if we did not authenticate yet but can
		if (!theCred.image() && theOwner->hasCredentials())
			return;
		// do not check acls if we sent invalid credentials
		if (theCred.image() && !theCred.valid())
			return;
	}

	checkAclMatch(acl.deny(), "denied");
}

void CltXact::checkAclMatch(const RegExGroup *matchedGrp, const char *action) {
	RegExMatchee m;
	buildAclMatchee(m);
	static Array<RegExGroup*> matches;
	matches.reset();
	theOwner->cfg()->acl().match(m, matches);

	bool explainMatch = false;

	if (matches.count() == 0)
		explainMatch = ReportError(errAclNoMatches);
	else
	if (matches.count() == 1 && matches.last() != matchedGrp)
		explainMatch = ReportError(errAclWrongMatch);
	else
	if (matches.count() > 1)
		explainMatch = ReportError(errAclManyMatches);

	if (explainMatch)
		explainAclMatch(m, action, matches);
}

void CltXact::buildAclMatchee(RegExMatchee &m) const {
	// form complete URL
	static WrBuf buf;
	buf.reset();

	ostrstream os(buf.space(), buf.spaceSize()-Size(1));
	m.urlHost = buf.space() + 0; // URL host alone
	Oid2UrlHost(theOid, os);
	Should(os << ends);
	
	m.urlPath = buf.space() + os.tellp(); // URL path alone
	Oid2UrlPath(theOid, os);
	Should(os << ends);
	
	m.url = buf.space() + os.tellp(); // complete URL
	Oid2Url(theOid, os);
	Should(os << ends);

	buf.appended(Size(os.tellp()));
	buf.append("", 1); // terminate even if stream is full

	m.userName = theCred.image().cstr();
	m.memberships = &theOwner->memberships();
}

void CltXact::explainAclMatch(const RegExMatchee &m, const char *action, const Array<RegExGroup*> &ourMatches) const {
	ostream &os = Comment(6) << "URL: " << m.url << endl;
	if (theCred.valid())
		os << "\tuser: " << theCred.image() << endl;
	os << "\tmembership maps: " << m.memberships->count() << endl;
	os << "\tgroups: ";
	for (int g = 0; g < m.memberships->count(); ++g) {
		if (g)
			os << "; ";
		dumpMatchingGroupNames(os, m.memberships->item(g));
		if (g >= 10) {
			os << "; ...";
			break;
		}
	}
	os << endl;
	os << "\twas probably " << action
		<< " but matches " << ourMatches.count() << " rule(s)";
	for (int i = 0; i < ourMatches.count(); ++i) {
		os << (i == 0 ? ": " : ", ");
		os << theOwner->cfg()->acl().ruleName(ourMatches[i]);
	}
	os << endc;
}

void CltXact::dumpMatchingGroupNames(ostream &os, const MembershipMap *map) const {
	MembershipMap::GroupIterator i = map->groupIterator(theCred);
	for (int count = 0; i && count <= 5; ++count, ++i) {
		if (count)
			os << ", ";
		os << *i;
	}
}

void CltXact::checkFreshness() {
	// cannot check without the date header
	if (theRepHdr.theDate < 0)
		return;

	// skip in-transit modicications to avoid false errors due to racing
	if (HttpDateAtMost(theStartTime) <= olcTimes().lmt())
		return;

	// response is stale if it was generated before modification time
	if (theRepHdr.theDate < olcTimes().lmt())
		noteError(errStaleHit);
}

bool CltXact::cfgAbortedReply() const {
	Assert(theRepSize.expected().known());
	Size newRepSize = theRepSize.expected();

	if (!theRepHdr.theAbortCoord)
		return false;

	const Size abSz =
		theRepHdr.theAbortCoord.pos(theRepHdr.theHdrSize, theRepSize.expected()-theRepHdr.theHdrSize);

	if (abSz < 0)
		return false;

	/* expecting abort, sooner or later */

	newRepSize = abSz;

	// we may have leftovers if we were parsing aborted content
	if (abSz <= theRepSize.actual() + unconsumed())
		return true; // reached abort level

	return false;
}

Error CltXact::setStatusCode(int aStatus) {
	theHttpStatus = aStatus;
	if (theHttpStatus != RepHdr::sc200_OK && 
		theHttpStatus != RepHdr::sc304_NotModified &&
		theHttpStatus != RepHdr::sc403_Forbidden &&
		theHttpStatus != RepHdr::sc407_ProxyAuthRequired &&
		!theRepHdr.redirect())
		return errHttpStatusCode;
	return Error();
}

void CltXact::checkDateSync() {
	if (theRepHdr.theDate < 0) {
		ReportError(errHttpNoDate);
		return;
	}

	const Time maxGap = Time::Sec(60);

	// cannot measure drift using replies that took too long
	if (TheClock - theStartTime >= maxGap/2)
		return;

	if (theRepHdr.theDate > TheClock.time() + maxGap ||
		theRepHdr.theDate < TheClock.time() - maxGap) {
		if (ReportError(errSyncDate)) {
			HttpDatePrint(Comment << "request generated:  ", theStartTime) << endc;
			HttpDatePrint(Comment << "response generated: ", theRepHdr.theDate) << endc;
			HttpDatePrint(Comment << "response received:  ", TheClock.time()) << endc;
			const Time diff = theRepHdr.theDate - TheClock.time();
			Comment << "time difference:  " << diff << endc;
		}
		return;
	}
}

void CltXact::updatePubWorld(const ObjWorld &slice) {
	PubWorld &pubWorld = *TheHostMap->at(theOid.viserv())->thePubWorld;
	int sliceIdx = -1;
	if (pubWorld.find(slice.id(), sliceIdx))
		pubWorld.sliceAt(sliceIdx).update(slice);
	else
		pubWorld.add(slice);
}

// called by BodyParsers, must not finish()
void CltXact::noteContent(const ParseBuffer &content) {
	if (theRepHdr.theChecksum.set())
		theCheckAlg.update(content.data(), content.size());
}

// called by BodyParsers, must not finish()
Error CltXact::noteEmbedded(ReqHdr &hdr) {
	TheEmbedStats.urlSeen++;

	const Error err = hdr.theUri.oid.foreignUrl() ?
		handleForeignEmbedOid(hdr) : handleEmbedOid(hdr);

	if (!err) {
		// "fix" new oid record
		hdr.theUri.oid.repeat(theOid.repeat());
		hdr.theUri.oid.hot(theOid.hot());
		hdr.theUri.oid.ims200(theOid.ims200());
		hdr.theUri.oid.ims304(theOid.ims304());
		hdr.theUri.oid.reload(theOid.reload());
		hdr.theUri.oid.get(true);
		theOwner->noteEmbedded(this, hdr.theUri.oid);
	}

	return err;
}

// called by BodyParsers, must not finish()
void CltXact::noteTrailerHeader(const ParseBuffer &hdr) {
	if (ReportError(errTrailerHeader))
		printMsg(hdr.data(), hdr.size());
}

// called by BodyParsers, must not finish()
void CltXact::noteEndOfTrailer() {
	Should(!theRepSize.expectingWhatParsed());
	theRepSize.expectingWhatParsed(true);
}

Error CltXact::handleEmbedOid(ReqHdr &hdr) {
	Error err;
	if (validRelOid(hdr.theUri.oid)) {
		if (hdr.theUri.host)
			err = setViserv(hdr.theUri.host, hdr.theUri.oid);
		else
			hdr.theUri.oid.viserv(theOid.viserv()); // relative URL

		if (!err)
			theOwner->selectTarget(hdr.theUri.oid);
	} else {
		err = errBadEmbedUri;
	}

	return err;
}

Error CltXact::handleForeignEmbedOid(ReqHdr &hdr) {
	Error err;
	//if (strncmp(hdr.theUri.pathBuf, "/cgi-bin/cntmgr.pl", 18) == 0)
	//	TheEmbedStats.scriptMgrUrlSeen++;
	
	if (!hdr.theUri.host) {
		// must set host name for Client to know where to send requests
		hdr.theUri.oid.viserv(theOid.viserv());
		const NetAddr &visName = TheHostMap->at(theOid.viserv())->theAddr;
		char buf[4*1024];
		ostrstream os(buf, sizeof(buf)-1);
		os << "http://" << visName.addrA() << ':' << visName.port() 
			<< hdr.theUri.oid.foreignUrl() << ends;
		buf[sizeof(buf)-1] = '\0';
		hdr.theUri.oid.foreignUrl(buf);
	}

	//if (TheEmbedStats.scriptUrlSeen % 1000 == 0) {
	//	(clog << here << "tag url: ").write(hdr.theUri.pathBuf, hdr.theUri.pathLen);
	//	clog << endl;
	//}

	// hdr.theUri.oid is already set
	return err;
}

Error CltXact::setViserv(const NetAddr &name, ObjId &oid) const {
	// XXX: merge with SrvXact::setViserv() ?
	int viserv = -1;
	if (!TheHostMap->find(name, viserv))
		return errForeignHostName;

	if (viserv != theOid.viserv())
		return errSrvRedirect; // disallow for now

	oid.viserv(viserv);
	return Error();
}

void CltXact::logStats(OLog &ol) const {
	Xaction::logStats(ol);
	ol << theOwner->seqvId();
	ol << (theSrvRep ? theSrvRep->serverIdx() : -1);
}
