
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include <ctype.h>

#include "xstd/Rnd.h"
#include "xstd/gadgets.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"
#include "runtime/polyErrors.h"
#include "client/CltXact.h"
#include "client/CltOpts.h"
#include "client/ParseBuffer.h"
#include "client/UriScriptBodyParser.h"

// commonly used RE patterns 
static const String ptnConst = "\"([^\"]*)\"";
static const String ptnName = "([[:alnum:]_]+)";
static const String ptnSpace = "[[:space:]]*";
static const String ptnEq = ptnSpace + "=" + ptnSpace;
static const String ptnPlus = ptnSpace + "\\+" + ptnSpace;

RegEx UriScriptBodyParser::TheFramelessFunRe;
RegEx UriScriptBodyParser::TheFramefullFunRe;
RegEx UriScriptBodyParser::TheUriRe;
RegEx UriScriptBodyParser::TheRspMsgRe;

BodyParserFarmT<UriScriptBodyParser> UriScriptBodyParser::TheParsers;


BodyParser *UriScriptBodyParser::GetOne(CltXact *anOwner, const CltCfg *aCfg) {
	if (!TheParsers.capacity())
		TheParsers.limit(1024);

	UriScriptBodyParser *parser = TheParsers.getTyped();
	parser->configure(anOwner, aCfg);
	return parser;
}

UriScriptBodyParser::UriScriptBodyParser() {

	// configure static REs if needed

	if (!TheFramelessFunRe.configured())
		ConfigureScriptFunRe(TheFramelessFunRe, "NotFrame_Message");

	if (!TheFramefullFunRe.configured())
		ConfigureScriptFunRe(TheFramefullFunRe, "Frame_Message");

	if (!TheUriRe.configured()) {
		String pattern;
		pattern = "[\"']";
		pattern += "(http://[^\"']+)";
		pattern += "[\"']";

		TheUriRe.configure("URI RE", pattern);
		Should(TheUriRe);
	}

	if (!TheRspMsgRe.configured()) {
		String pattern;
		pattern = "M_Rsp_Message\\.src";
		pattern += ptnEq;
		pattern += ptnConst;
		pattern += ptnPlus;
		pattern += ptnName;
		pattern += ptnPlus;
		pattern += ptnConst;
		pattern += ptnPlus;
		pattern += ptnName;
		pattern += ";";

		TheRspMsgRe.configure("M_Rsp_Message.src RE", pattern);
		Should(TheRspMsgRe);
	}

	resetSelf();
}

void UriScriptBodyParser::reset() {
	resetSelf();
	BodyParser::reset();
}

void UriScriptBodyParser::resetSelf() {
	theCfg = 0;
	theScriptBeg = theScriptEnd = 0;
	theFrameFunRe = 0;
	theRspMsgIdx = -1;
	theState = stNone;
}

void UriScriptBodyParser::configure(CltXact *anOwner, const CltCfg *aCfg) {
	BodyParser::configure(anOwner);
	Check(!theCfg && aCfg);
	theCfg = aCfg;

	// XXX: kludge to do 95%/5% split
	static RndGen rng;
	if (rng.event(0.95)) {
		theFrameFunRe = &TheFramelessFunRe;
		theRspMsgIdx = 0;
	} else {
		theFrameFunRe = &TheFramefullFunRe;
		theRspMsgIdx = 1;
	}

	Should(TheRspMsgRe.configured());
}

BodyParserFarm &UriScriptBodyParser::farm() const {
	return TheParsers;
}

Size UriScriptBodyParser::parse(const ParseBuffer &buf) {
	if (theState == stNone)
		theState = stOpen;

	if (theState == stOpen)
		parseOpen(buf);

	if (theState == stClose)
		parseClose(buf);

	if (theState == stBody)
		parseBody(buf);

	if (theState == stDone)
		return parseAny(buf);

	return 0; // need more data
}

void UriScriptBodyParser::parseOpen(const ParseBuffer &buf) {
	static const String pfx = "<script";
	if (buf.size() < pfx.len())
		return; // need more data

	if (strncmp(buf.data(), pfx.cstr(), pfx.len()) == 0)
		theState = stClose;
	else
		theState = stDone;
}

void UriScriptBodyParser::parseClose(const ParseBuffer &buf) {
	static const String sfx = "</script";
	if (buf.size() < sfx.len())
		return; // need more data
	
	const char *bufEnd = buf.data() + buf.size();
	theScriptBeg = buf.data();
	theScriptEnd = StrBoundStr(theScriptBeg, sfx.cstr(), bufEnd);
	if (theScriptEnd)
		theState = stBody;
}

void UriScriptBodyParser::parseBody(const ParseBuffer &buf) {
	if (Should(theScriptBeg) && Should(theScriptEnd)) {
		parseFrameFun();
		parseRspMsg();
	}
	theState = stDone;
}

void UriScriptBodyParser::parseFrameFun() {
	Assert(theFrameFunRe);
	static RegEx::Matches matches(1+1);
	const char *p = theScriptBeg;
	if (findNextRe(*theFrameFunRe, matches, p)) {
		const RegEx::Match &m = matches[1];
		parseUris(theScriptBeg + m.rm_so, theScriptBeg + m.rm_eo);
	}
else cerr << here << "no " << theFrameFunRe->image() << endl;
}

void UriScriptBodyParser::parseUris(const char *beg, const char *end) {
	static RegEx::Matches matches(1+1);
	const char *p = beg;
	while (findNextRe(TheUriRe, matches, p, end)) {
		const RegEx::Match &m = matches[1];
		const String uri(theScriptBeg + m.rm_so,  m.rm_eo - m.rm_so);
		forwardUri(uri);
	}
}
	
void UriScriptBodyParser::parseRspMsg() {
	static RegEx::Matches matches(1+4);
	const char *p = theScriptBeg;
	for (int i = 0; i <= theRspMsgIdx && findNextRe(TheRspMsgRe, matches, p); ++i) {
		if (i != theRspMsgIdx)
			continue;

		String uri;
		const bool ok = parseConstValue(uri, matches[1]) &&
			parseNameValue(uri, matches[2]) &&
			parseConstValue(uri, matches[3]) &&
			parseNameValue(uri, matches[4]);

		if (ok)
			forwardUri(uri); 
		// else, individual parse* routines already complained
	}
}

void UriScriptBodyParser::forwardUri(const String &uri) {
	Error err;
	const char *parsep = uri.cstr();
	ReqHdr hdr;
	if (hdr.parseUri(parsep, parsep+uri.len(), hdr.theUri))
		err = theOwner->noteEmbedded(hdr);
	else
		err = errForeignTag;

	if (err) {
		if (!TheCltOpts.ignoreBadContTags && ReportError(err)) {
			dumpContext(Comment << "unparseable recreated URL: ",
				uri.data(), uri.len()) << endc;
		}
	} else {
		static int reportCount = 0;
		if (!reportCount++) {
			dumpContext(Comment << "fyi: first URL extracted " <<
				"from a script: ", uri.data(), uri.len()) << endc;
		}
	}
}

bool UriScriptBodyParser::findNextRe(const RegEx &re, RegEx::Matches &matches, const char *&start, const char *end) {
	if (!end)
		end = theScriptEnd;
	const int expectedCount = matches.capacity();
	matches.reset();
	matches.append(RegEx::StartEnd(start-theScriptBeg, end-start));
	if (re.match(theScriptBeg, matches, RegEx::reStartEnd) && 
		Should(matches.count() == expectedCount)) {
		start = theScriptBeg + matches[0].rm_eo; // advance iteration pos
		return true;
	}
	return false;
}
	

bool UriScriptBodyParser::parseConstValue(String &uri, const RegEx::Match &m) {
	if (Should(-1 < m.rm_so && m.rm_so <= m.rm_eo)) {
		uri.append(theScriptBeg + m.rm_so, m.rm_eo - m.rm_so);
		return true;
	}
	return false;
}

bool UriScriptBodyParser::parseNameValue(String &uri, const RegEx::Match &nm) {
	if (!Should(-1 < nm.rm_so && nm.rm_so < nm.rm_eo))
		return false;

	// possible optimization: search for each variable definition only once

	const String name(theScriptBeg + nm.rm_so, nm.rm_eo - nm.rm_so);
	RegEx *re = 0;
	if (name == "IP_Message") {
		static RegEx reIp;
		re = &reIp;
	} else
	if (name == "GUID_Message") {
		static RegEx reGuid;
		re = &reGuid;
	} else {
		const Error &err = errForeignTag;
		if (!TheCltOpts.ignoreBadContTags && ReportError(err)) {
			dumpContext(Comment << "unknown variable name in URI formula: ",
				theScriptBeg + nm.rm_so, nm.rm_eo - nm.rm_so) << endc;
		}
		return false;
	}

	return Should(re) && findNameValue(uri, *re, name);
}

bool UriScriptBodyParser::findNameValue(String &uri, RegEx &re, const String &name) {
	if (re.image().len() <= 0) {
		String pattern = name;
		pattern += ptnEq;
		pattern += ptnConst;
		re.configure(name + " RE", pattern);
		Should(re);
	}

	RegEx::Matches matches(1+1);
	matches.append(RegEx::StartEnd(0, theScriptEnd - theScriptBeg));
	if (re.match(theScriptBeg, matches, RegEx::reStartEnd) &&
		Should(matches.count() == 2))
		return parseConstValue(uri, matches[1]);
	
	const Error &err = errForeignTag;
	if (!TheCltOpts.ignoreBadContTags && ReportError(err)) {
		dumpContext(Comment << "undefined variable name in URI formula: ",
			name.data(), name.len()) << endc;
		dumpContext(Comment << "script with the above variable: " << endl,
			theScriptBeg, theScriptEnd - theScriptBeg) << endc;
	}
	return false;
}

Size UriScriptBodyParser::parseAny(const ParseBuffer &buf) {
	const Size parsedSize = buf.size();
	theOwner->noteContent(buf);
	return parsedSize;
}

void UriScriptBodyParser::noteLeftovers(const ParseBuffer &leftovers) {
	Should(theState == stOpen || theState == stClose);
	if (theState == stClose && ReportError(errContentLeftovers)) {
		const char *what = theState == stOpen ? "beginning" : "end";
		dumpContext(Comment << "unable to locate the " << what <<
			" of a URI-setting <script> in the remaining " <<
			leftovers.size() << " of content leftovers near " << endl,
			leftovers.data(), leftovers.size()) << endc;
	}
	theOwner->noteContent(leftovers);
}

void UriScriptBodyParser::noteOverflow(const ParseBuffer &buf) {
	if (ReportError(errHugeContentToken)) {
		dumpContext(Comment << "huge URI-setting <script> context near ",
		buf.data(), buf.size()) << endc;
	}
	theOwner->noteContent(buf);
}


void UriScriptBodyParser::ConfigureScriptFunRe(RegEx &re, const String &name) {
	String pattern;
	pattern = "function";
	pattern += ptnSpace;
	pattern += name;
	pattern += "\\(\\)";
	pattern += ptnSpace;
	pattern += "{";
	pattern += "([^}]*)"; // body
	pattern += "}";

	re.configure(name + "() RE", pattern);
	Should(re);
}
