
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#ifndef X_RING_H
#define X_RING_H

#include "xstd/Array.h"
#include "xstd/Rnd.h"

// FIFO Queue of <Item>s with fixed capacity (i.e., ring)
// note dynamic resize()-ing on enqueue() does work but is disabled

template <class Item>
class Ring: protected Array<Item> {
	public:
		Ring(int aCapacity = 0): Array<Item>(aCapacity), theInOff(0), theOutOff(0) {}

		int capacity() const { return this->theCapacity; }
		int count() const { return theInOff - theOutOff; }
		int size() const { return Array<Item>::size(); }
		bool empty() const { return theInOff <= theOutOff; }
		bool full() const { return count() >= this->theCapacity; }
		const Item &top(int off = 0) const { return item((theOutOff+off) % this->theCapacity); }

		Item &top(int off = 0) { return item((theOutOff+off) % this->theCapacity); }
		void enqueue(Item i) { Assert(!full()); item(theInOff++ % this->theCapacity) = i; check(); }
		Item dequeue() { return item(theOutOff++ % this->theCapacity); }

		void reset() { theInOff = theOutOff = 0; }
		inline void resize(int aCap);

		inline void randomize(RndGen &rng);

	protected:
		inline void check();

	protected:
		int theInOff;
		int theOutOff;
};


/* implementaion of in-lined methods */

template <class Item>
inline
void Ring<Item>::randomize(RndGen &rng) {
	for (int i = theOutOff; i < theInOff; ++i)
		Array<Item>::swap(i % this->theCapacity,
			rng(theOutOff, theInOff) % this->theCapacity);
}

template <class Item>
inline
void Ring<Item>::resize(int aCap) {
	const int oldCap = this->theCapacity;
	Item *oldItems = this->theItems;
	this->theItems = 0;

	this->grow(aCap);

	if (oldItems) {
		if (count()) {
			const int old_hi = theInOff > oldCap ? oldCap : theInOff;
			const int gap = this->theCapacity - old_hi;
			const int new_low = theOutOff + gap;
			if (old_hi > theOutOff)
				memmove(new_low, oldItems+theOutOff, old_hi-theOutOff);
			if (theInOff > oldCap)
				memmove(0, oldItems, theInOff-oldCap);
			theInOff += gap;
			theOutOff += gap;
		}
		delete[] oldItems;
	}
}

template <class Item>
inline
void Ring<Item>::check() {
	Assert(theInOff > theOutOff && this->count() <= this->theCapacity);
	if (theOutOff >= this->theCapacity) {
		theInOff %= this->theCapacity;
		theOutOff %= this->theCapacity;
		if (theInOff <= this->theOutOff)
			theInOff += this->theCapacity;
	}
}

#endif
